"""Peer-to-peer protocol for voice assistants."""

from .version import __version__

__author__ = "Michael Hansen"
__email__ = "mike@rhasspy.org"

__all__ = ["__author__", "__email__", "__version__"]
